import pytest
from ruamel.yaml import YAML

from somesy.mkdocs.writer import MkDocs


def test_mkdocs_validate_accept(load_files, file_types):
    """Validate by loading the data file using the fixture."""
    load_files([file_types.MKDOCS])


def test_mkdocs_validate_reject(tmp_path):
    """Test validating a mkdocs file."""
    mkdocs_path = tmp_path / "mkdocs.yml"

    # create a mkdocs file with a invalid version
    reject_mkdocs_object = {"site_description": "A project"}
    yaml = YAML()
    yaml.dump(reject_mkdocs_object, mkdocs_path)

    # try to load the mkdocs file
    with pytest.raises(ValueError):
        MkDocs(mkdocs_path)



"""Writer adapter for pom.xml files."""
import logging
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from somesy.core.models import Person
from somesy.core.writer import FieldKeyMapping, ProjectMetadataWriter

from . import POM_ROOT_ATRS, POM_URL
from .xmlproxy import XMLProxy

logger = logging.getLogger("somesy")


class POM(ProjectMetadataWriter):
    """Java Maven pom.xml parser and saver."""

    # TODO: write a wrapper for ElementTree that behaves like a dict
    # TODO: set up correct field name mappings

    def __init__(
        self,
        path: Path,
        create_if_not_exists: bool = True,
    ):
        """Java Maven pom.xml parser.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        mappings: FieldKeyMapping = {
            # "year": ["inceptionYear"],  # not supported by somesy + does not really change
            # "project_slug": ["artifactId"],  # not supported by somesy for sync
            "license": ["licenses", "license"],
            "homepage": ["url"],
            "repository": ["scm"],
            "documentation": ["distributionManagement", "site"],
            "authors": ["developers", "developer"],
            "contributors": ["contributors", "contributor"],
        }
        super().__init__(
            path, create_if_not_exists=create_if_not_exists, direct_mappings=mappings
        )

    def _init_new_file(self):
        """Initialize new pom.xml file."""
        pom = XMLProxy(ET.Element("project", POM_ROOT_ATRS))
        pom["properties"] = {"info.versionScheme": "semver-spec"}
        pom.write(self.path)

    def _load(self):
        """Load the POM file."""
        ET.register_namespace("", POM_URL)  # register POM as default xml namespace
        self._data = XMLProxy.parse(self.path, default_namespace=POM_URL)

    def _validate(self):
        """Validate the POM file."""
        logger.info("Cannot validate POM file, skipping validation.")

    def save(self, path: Optional[Path] = None) -> None:
        """Save the POM DOM to a file."""
        self._data.write(path or self.path, default_namespace=None)

    def _get_property(
        self,
        key: Union[str, List[str]],
        *,
        only_first: bool = False,
        remove: bool = False,
    ) -> Optional[Any]:
        elem = super()._get_property(key, only_first=only_first, remove=remove)
        if elem is not None:
            if isinstance(elem, list):
                return [e.to_jsonlike() for e in elem]
            else:
                return elem.to_jsonlike()
        return None

    @staticmethod
    def _from_person(person: Person):
        """Convert person object to dict for POM XML person format."""
        ret: Dict[str, Any] = {}
        person_id = str(person.orcid) or person.to_name_email_string()
        ret["id"] = person_id
        ret["name"] = person.full_name
        ret["email"] = person.email
        if person.orcid:
            ret["url"] = str(person.orcid)
        if person.contribution_types:
            ret["roles"] = dict(role=[c.value for c in person.contribution_types])
        return ret

    @staticmethod
    def _to_person(person_obj) -> Person:
        """Parse POM XML person to a somesy Person."""
        print(person_obj)
        names = person_obj["name"].split()
        gnames = " ".join(names[:-1])
        fname = names[-1]
        email = person_obj["email"]
        url = person_obj.get("url")
        maybe_orcid = url if url.find("orcid.org") >= 0 else None
        if roles := person_obj.get("roles"):
            contr = roles["role"]
        else:
            contr = None

        return Person(
            given_names=gnames,
            family_names=fname,
            email=email,
            orcid=maybe_orcid,
            contribution_types=contr,
        )

    # no search keywords supported in POM
    @property
    def keywords(self) -> Optional[List[str]]:
        """Return the keywords of the project."""
        pass

    @keywords.setter
    def keywords(self, keywords: List[str]) -> None:
        """Set the keywords of the project."""
        pass

    # authors must be a list
    @property
    def authors(self):
        """Return the authors of the project."""
        authors = self._get_property(self._get_key("authors"))
        return authors if isinstance(authors, list) else [authors]

    @authors.setter
    def authors(self, authors: List[Person]) -> None:
        """Set the authors of the project."""
        authors = [self._from_person(c) for c in authors]
        self._set_property(self._get_key("authors"), authors)

    # contributors must be a list
    @property
    def contributors(self):
        """Return the contributors of the project."""
        contr = self._get_property(self._get_key("contributors"))
        if contr is None:
            return []
        return contr if isinstance(contr, list) else [contr]

    @contributors.setter
    def contributors(self, contributors: List[Person]) -> None:
        """Set the contributors of the project."""
        contr = [self._from_person(c) for c in contributors]
        self._set_property(self._get_key("contributors"), contr)

    # no maintainers supported im POM, only developers and contributors
    @property
    def maintainers(self):
        """Return the maintainers of the project."""
        return []

    @maintainers.setter
    def maintainers(self, maintainers: List[Person]) -> None:
        """Set the maintainers of the project."""
        pass

    # only one project license supported in somesy (POM can have many)
    @property
    def license(self) -> Optional[str]:
        """Return the license of the project."""
        lic = self._get_property(self._get_key("license"), only_first=True)
        return lic.get("name") if lic is not None else None

    @license.setter
    def license(self, license: Optional[str]) -> None:
        """Set the license of the project."""
        self._set_property(
            self._get_key("license"), dict(name=license, distribution="repo")
        )

    @property
    def repository(self) -> Optional[Union[str, dict]]:
        """Return the repository url of the project."""
        repo = super().repository
        if isinstance(repo, str):
            return repo
        return repo.get("url") if repo is not None else None

    @repository.setter
    def repository(self, value: Optional[Union[str, dict]]) -> None:
        """Set the repository url of the project."""
        self._set_property(
            self._get_key("repository"), dict(name="git repository", url=value)
        )

    @property
    def documentation(self) -> Optional[Union[str, dict]]:
        """Return the documentation url of the project."""
        docs = super().documentation
        if isinstance(docs, str):
            return docs
        return docs.get("url") if docs is not None else None

    @documentation.setter
    def documentation(self, value: Optional[Union[str, dict]]) -> None:
        """Set the documentation url of the project."""
        self._set_property(
            self._get_key("documentation"), dict(name="documentation site", url=value)
        )

    def sync(self, metadata) -> None:
        """Sync codemeta.json with project metadata.

        Use existing sync function from ProjectMetadataWriter but update repository and contributors.
        """
        super().sync(metadata)
        self.contributors = self._sync_person_list(self.contributors, metadata.people)



from somesy.pom_xml import POM_URL
from somesy.pom_xml.xmlproxy import XMLProxy

import pytest


def test_parse_write(tmp_path, xml_examples):
    """Make sure that header + namespace of XML are set correctly and comments are preserved."""
    file_src = xml_examples("blank_pom.xml")
    file_trg = tmp_path / "written.xml"

    prx = XMLProxy.parse(file_src, default_namespace=POM_URL)

    # the root
    assert repr(prx).startswith(
        "<Element '{http://maven.apache.org/POM/4.0.0}project' at"
    )

    # the comment element
    assert len(prx) == 1
    lst = list(iter(prx))
    assert len(lst) == 1 and repr(lst[0]).startswith("<Element <function Comment")

    # check that namespaceing and comments are preserved
    prx.write(file_trg)
    contents = (file_trg).read_text()
    assert contents.strip() == file_src.read_text().strip()


def test_read_access(xml_examples):
    prx = XMLProxy.parse(xml_examples("example_1.xml"), default_namespace=POM_URL)

    # __iter__, __len__, __bool__, __repr__
    assert str(prx).startswith("<Element")
    assert len(prx) == 9
    assert bool(prx)
    elems = list(iter(prx))
    assert len(elems) == 9
    assert not elems[2].is_comment
    assert elems[2].tag == "listEntry"
    assert not len(elems[1])
    assert not bool(elems[1])
    assert elems[3].is_comment
    assert elems[3].tag is None

    with pytest.raises(ValueError):
        elems[3].tag = "newTag"

    el = prx["anInt"]
    el.tag = "someInt"  # change
    assert "anInt" not in prx
    assert "someInt" in prx
    el.tag = "anInt"  # restore

    # __contains__
    assert "nothing" in prx
    assert "anInt" in prx
    assert "listEntry" in prx
    assert "invalid" not in prx

    # __getitem__ / get
    with pytest.raises(KeyError):
        assert prx["invalid"]
    assert prx.get("invalid") is None
    assert len(prx.get("invalid", as_nodes=True)) == 0

    # self-closing tag and empty open/close are both an empty string
    assert prx["nothing"].tag == "nothing"
    assert prx["nothing"].to_jsonlike() == ""
    assert prx["emptyString"].to_jsonlike() == ""

    # primitive values returned as strings
    assert prx["anInt"].to_jsonlike() == "42"
    assert prx["aBool"].to_jsonlike() == "true"

    # multiple elements -> returns list
    lst = prx["listEntry"]
    assert len(lst) == 2
    assert lst[0].to_jsonlike() == "foo"
    assert lst[1].to_jsonlike() == {"someDict": {"a": "x", "b": "y"}, "someValue": "z"}

    expected = dict(
        nothing="",
        emptyString="",
        listEntry=["foo", {"someDict": {"a": "x", "b": "y"}, "someValue": "z"}],
        __comment_1__=" comment 1 ",
        aBool="true",
        __comment_2__=" comment 2 ",
        anInt="42",
        aFloat="3.14",
    )
    dct = prx.to_jsonlike()
    assert dct == expected
    assert list(dct.keys()) == list(expected.keys())  # key order correct


def test_del_set_items(xml_examples):
    prx = XMLProxy.parse(xml_examples("example_1.xml"), default_namespace=POM_URL)

    # remove existing elements
    del prx["emptyString"]
    del prx["listEntry"][1]["someDict"]["b"]

    with pytest.raises(KeyError):
        del prx["emptyString"]  # already removed
    with pytest.raises(KeyError):
        del prx["invalid"]  # not existing

    # add new elements
    prx["newValue"] = 5
    assert prx["newValue"].to_jsonlike() == "5"
    dct = dict(some="hello", keys="world")
    prx["newObject"] = dct
    assert prx["newObject"].to_jsonlike() == dct

    # overwrite existing elements
    prx["anInt"] = 123  # primitive to primitive
    assert prx["anInt"].to_jsonlike() == "123"
    prx["newValue"] = dict(x=1, y=2)  # primitive to complex
    assert prx["newValue"].to_jsonlike() == dict(x="1", y="2")
    prx["newValue"] = None  # complex to primitive
    assert prx["newValue"].to_jsonlike() == ""
    prx["listEntry"][1]["someDict"]["a"] = "qux"  # overwrite nested
    assert prx["listEntry"][1]["someDict"]["a"].to_jsonlike() == "qux"

    # overwrite specific if there are many hits
    prx[prx["listEntry"][0]] = "xyz"
    assert prx["listEntry"][0].to_jsonlike() == "xyz"
    prx[prx["listEntry"][0]] = dict(x="1", y="2")
    assert prx["listEntry"][0].to_jsonlike() == dict(x="1", y="2")
    # overwrite all elements if there are many hits
    prx["listEntry"] = 321
    assert prx["listEntry"].to_jsonlike() == "321"
    # overwrite all elements if there are many target values
    prx["listEntry"] = [123, 456]
    assert prx.get("listEntry", deep=True) == ["123", "456"]

    # elements still in right order
    exp_keys = ["nothing", None, "aBool", None, "anInt", "aFloat"]
    exp_keys += ["newValue", "newObject", "listEntry", "listEntry"]
    keys = [x.tag for x in iter(prx)]
    assert keys == exp_keys


def test_update_write(xml_examples, tmp_path):
    prx = XMLProxy.parse(xml_examples("example_pom.xml"), default_namespace=POM_URL)

    prx["version"] = "0.2.0"
    prx["licenses"]["license"] = "MIT"
    prx["developers"]["developer"][1]["name"] = "John Doe"
    prx["newKey"] = 42
    externalKey = "{https://something.com}externalKey"
    prx[externalKey] = "value"

    tmp_file = tmp_path / "written.xml"
    prx.write(tmp_file)

    # check that the comment is retained
    content = tmp_file.read_text()
    assert content.find("some comment") > 0
    # check that the changes persisted
    prs = XMLProxy.parse(tmp_file, default_namespace=POM_URL)
    assert prs["developers"]["developer"][1]["name"].to_jsonlike() == "John Doe"
    assert prs["version"].to_jsonlike() == "0.2.0"
    assert "newKey" in prs
    # check that the namespace of the external key is preserved
    assert "externalKey" not in prs
    assert externalKey in prs



import pytest

from somesy.mkdocs import MkDocs
from somesy.core.models import Person


@pytest.fixture
def mkdocs(load_files, file_types):
    files = load_files([file_types.MKDOCS])
    return files[file_types.MKDOCS]


def test_content_match(mkdocs: MkDocs):
    assert mkdocs.name == "test-package"
    assert mkdocs.description == "This is a test package for demonstration purposes."
    assert len(mkdocs.authors) == 0


def test_sync(mkdocs: MkDocs, somesy_input: dict):
    mkdocs.sync(somesy_input.project)
    assert mkdocs.name == "testproject"
    assert mkdocs.authors[0] == "John Doe <john.doe@example.com>"
    assert mkdocs.repo_name == "/example/testproject"


def test_save(tmp_path):
    # test save with default path
    file_path = tmp_path / "mkdocs.yml"
    mkdocs = MkDocs(file_path, create_if_not_exists=True)
    mkdocs.save()
    assert file_path.is_file()

    # test save with custom path
    custom_path = tmp_path / "mkdocs2.yml"
    mkdocs.save(custom_path)
    assert custom_path.is_file()


def test_from_person(person: Person):
    p = MkDocs._to_person(MkDocs._from_person(person))
    assert p.to_name_email_string() == person.to_name_email_string()



"""Julia model."""
import uuid
from typing import Optional, Set

from packaging.version import parse as parse_version
from pydantic import (
    BaseModel,
    EmailStr,
    Field,
    TypeAdapter,
    field_validator,
)
from typing_extensions import Annotated

EMailAddress = TypeAdapter(EmailStr)


class JuliaConfig(BaseModel):
    """Julia configuration model."""

    model_config = dict(use_enum_values=True)

    name: Annotated[
        str,
        Field(description="Package name"),
    ]
    version: Annotated[
        str,
        Field(
            pattern=r"^\d+(\.\d+)*((a|b|rc)\d+)?(post\d+)?(dev\d+)?$",
            description="Package version",
        ),
    ]
    uuid: Annotated[str, Field(description="Package UUID")]
    authors: Annotated[Optional[Set[str]], Field(description="Package authors")] = None

    @field_validator("version")
    @classmethod
    def validate_version(cls, v):
        """Validate version using PEP 440."""
        try:
            _ = parse_version(v)
        except ValueError as err:
            raise ValueError("Invalid version") from err
        return v

    @field_validator("authors")
    @classmethod
    def validate_email_format(cls, v):
        """Validate email format."""
        for author in v:
            if (
                not isinstance(author, str)
                or " " not in author
                or not EMailAddress.validate_python(author.split(" ")[-1][1:-1])
            ):
                raise ValueError("Invalid email format")
        return v

    @field_validator("uuid")
    @classmethod
    def validate_uuid(cls, v):
        """Validate uuid field."""
        try:
            _ = uuid.UUID(v)
        except ValueError as err:
            raise ValueError("Invalid UUID") from err
        return v



"""Set config files for somesy."""
import logging
from pathlib import Path

import typer

from somesy.commands import init_config
from somesy.core.core import discover_input
from somesy.core.log import SomesyLogLevel, set_log_level

from .util import wrap_exceptions

logger = logging.getLogger("somesy")
app = typer.Typer()


@app.command()
@wrap_exceptions
def config():
    """Set CLI configs for somesy."""
    # check if input file exists, if not, try to find it from default list
    input_file_default = discover_input()

    # prompt for inputs
    input_file = typer.prompt("Input file path", default=input_file_default)
    options = dict(input_file=Path(input_file))

    # ----

    options["no_sync_cff"] = not typer.confirm(
        "Do you want to sync to a CFF file?", default=True
    )
    if cff_file := typer.prompt("CFF file path", default="CITATION.cff"):
        options["cff_file"] = cff_file

    options["no_sync_codemeta"] = not typer.confirm(
        "Do you want to sync to a codemeta.json file?", default=True
    )
    if codemeta_file := typer.prompt(
        "codemeta.json file path", default="codemeta.json"
    ):
        options["codemeta_file"] = codemeta_file

    options["no_sync_pyproject"] = not typer.confirm(
        "Do you want to sync to a pyproject.toml file?", default=True
    )
    if pyproject_file := typer.prompt(
        "pyproject.toml file path", default="pyproject.toml"
    ):
        options["pyproject_file"] = pyproject_file

    options["sync_package_json"] = typer.confirm(
        "Do you want to sync to a package.json file?", default=False
    )
    if package_json_file := typer.prompt(
        "package.json file path", default="package.json"
    ):
        options["package_json_file"] = package_json_file

    options["no_sync_julia"] = not typer.confirm(
        "Do you want to sync to a Project.toml(Julia) file?", default=True
    )
    if julia_file := typer.prompt(
        "Project.toml (Julia) file path", default="Project.toml"
    ):
        options["julia_file"] = julia_file

    options["no_sync_fortran"] = not typer.confirm(
        "Do you want to sync to a fpm.toml(fortran) file?", default=True
    )
    fortran_file = typer.prompt("fpm.toml(fortran) file path", default="fpm.toml")
    if fortran_file is not None or fortran_file != "":
        options["fortran_file"] = fortran_file

    options["no_sync_pom_xml"] = not typer.confirm(
        "Do you want to sync to a pom.xml file?", default=True
    )
    if pom_xml_file := typer.prompt("pom.xml file path", default="pom.xml"):
        options["pom_xml_file"] = pom_xml_file

    options["no_sync_mkdocs"] = not typer.confirm(
        "Do you want to sync to a mkdocs.yml file?", default=True
    )
    if mkdocs_file := typer.prompt("mkdocs.yml file path", default="mkdocs.yml"):
        options["mkdocs_file"] = mkdocs_file

    options["no_sync_rust"] = not typer.confirm(
        "Do you want to sync to a Cargo.toml file?", default=True
    )
    if rust_file := typer.prompt("Cargo.toml file path", default="Cargo.toml"):
        options["rust_file"] = rust_file

    # ----

    options["show_info"] = typer.confirm(
        "Do you want to show info about the sync process?"
    )
    options["verbose"] = typer.confirm("Do you want to show verbose logs?")
    options["debug"] = typer.confirm("Do you want to show debug logs?")

    set_log_level(
        SomesyLogLevel.from_flags(
            debug=options["debug"],
            verbose=options["verbose"],
            info=options["show_info"],
        )
    )

    logger.debug(f"CLI options entered: {options}")

    init_config(input_file, options)
    logger.info(
        f"[bold green]Input file is updated/created at {input_file}[/bold green]"
    )



from pathlib import Path

import pytest

from somesy.core.models import LicenseEnum, Person, ProjectMetadata
from somesy.rust import Rust


@pytest.fixture
def rust(load_files, file_types):
    files = load_files([file_types.RUST])
    return files[file_types.RUST]


@pytest.fixture
def rust_file(create_files, file_types):
    folder = create_files([(file_types.RUST, "Cargo.toml")])
    return folder / Path("Cargo.toml")


def test_content_match(rust):
    # create a function to check both file formats
    assert rust.name == "test-package"
    assert rust.description == "This is a test package for demonstration purposes."
    assert rust.license == "MIT" or rust.license["text"] == "MIT"
    assert len(rust.authors) == 1


def test_sync(rust, somesy_input):
    rust.sync(somesy_input.project)
    assert rust.name == "testproject"
    assert rust.version == "1.0.0"


def test_save(tmp_path, rust):
    custom_path = tmp_path / Path("Cargo.toml")
    rust.save(custom_path)
    assert custom_path.is_file()
    custom_path.unlink()


def test_from_to_person(person):
    assert Rust._from_person(person) == f"{person.full_name} <{person.email}>"

    p = Rust._to_person(Rust._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email

    # rust also has only name format
    assert Rust._to_person(person.full_name) == None


def test_person_merge(rust_file, person):
    pj = Rust(rust_file)
    # update project file with known data
    pm = ProjectMetadata(
        name="My awesome project",
        description="Project description",
        license=LicenseEnum.MIT,
        version="0.1.0",
        people=[person.model_copy(update=dict(author=True, publication_author=True))],
    )
    pj.sync(pm)
    pj.save()
    # ----

    # jane becomes john -> modified person
    person1b = person.model_copy(
        update={"given_names": "John", "author": True, "publication_author": True}
    )

    # different Jane Doe with different orcid -> new person
    person2 = person.model_copy(
        update={
            "orcid": "https://orcid.org/4321-0987-3231",
            "email": "i.am.jane@doe.com",
            "author": True,
            "publication_author": True,
        }
    )
    # use different order, just for some difference
    person2.set_key_order(["given_names", "orcid", "family_names", "email"])

    # listed in "arbitrary" order in somesy metadata (new person comes first)
    pm.people = [person2, person1b]  # need to assign like that to keep _key_order
    pj.sync(pm)
    pj.save()

    # existing author info preserved, order not preserved because no orcid
    person1b_rep = Rust._from_person(person1b)
    person2_rep = Rust._from_person(person2)
    assert (pj.authors[0] == person1b_rep) or (pj.authors[1] == person1b_rep)
    assert (pj.authors[0] == person2_rep) or (pj.authors[1] == person2_rep)

    # new person
    person3 = Person(
        **{
            "given_names": "Janice",
            "family_names": "Doethan",
            "email": "jane93@gmail.com",
            "author": True,
            "publication_author": True,
        }
    )
    person3_rep = Rust._from_person(person3)

    # john has a new email address
    person1c = person1b.model_copy(update={"email": "john.of.us@qualityland.com"})
    person1c_rep = Rust._from_person(person1c)

    # jane 2 is removed from authors, but added to maintainers
    person2.author = False
    person2.publication_author = False
    person2.maintainer = True
    # reflect in project metadata
    pm.people = [person3, person2, person1c]
    # sync to CFF file
    pj.sync(pm)
    pj.save()

    assert pj.authors[0] == person1c_rep
    assert pj.authors[1] == person3_rep


def test_person_merge_no_email(rust_file, person):
    # try with a rust file that has an author with only name
    r = Rust(rust_file)
    r._data["package"]["authors"] = ["Jane Doe"]
    r.save()

    # update project file with known data
    pm = ProjectMetadata(
        name="My-awesome-project",
        description="Project description",
        license=LicenseEnum.MIT,
        version="0.1.0",
        people=[person.model_copy(update=dict(author=True, publication_author=True))],
    )

    r.sync(pm)
    assert r.authors[0] == Rust._from_person(person)



"""Sync selected metadata files with given input file."""
import logging
from pathlib import Path
from typing import Type

from rich.pretty import pretty_repr

from somesy.cff.writer import CFF
from somesy.codemeta import CodeMeta
from somesy.core.models import ProjectMetadata, SomesyInput
from somesy.core.writer import ProjectMetadataWriter
from somesy.fortran.writer import Fortran
from somesy.julia.writer import Julia
from somesy.mkdocs import MkDocs
from somesy.package_json.writer import PackageJSON
from somesy.pom_xml.writer import POM
from somesy.pyproject.writer import Pyproject
from somesy.rust import Rust

logger = logging.getLogger("somesy")


def _sync_file(
    metadata: ProjectMetadata, file: Path, writer_cls: Type[ProjectMetadataWriter]
):
    """Sync metadata to a file using the provided writer."""
    logger.verbose(f"Loading '{file.name}' ...")
    writer = writer_cls(file)
    logger.verbose(f"Syncing '{file.name}' ...")
    writer.sync(metadata)
    writer.save(file)
    logger.verbose(f"Saved synced '{file.name}'.\n")


def sync(somesy_input: SomesyInput):
    """Sync selected metadata files with given input file."""
    conf, metadata = somesy_input.config, somesy_input.project

    pp_metadata = pretty_repr(metadata.model_dump(exclude_defaults=True))
    logger.debug(f"Project metadata: {pp_metadata}")

    # update these only if they exist:

    if conf.pyproject_file.is_file() and not conf.no_sync_pyproject:
        _sync_file(metadata, conf.pyproject_file, Pyproject)

    if conf.package_json_file.is_file() and not conf.no_sync_package_json:
        _sync_file(metadata, conf.package_json_file, PackageJSON)

    if conf.julia_file.is_file() and not conf.no_sync_julia:
        _sync_file(metadata, conf.julia_file, Julia)

    if conf.fortran_file.is_file() and not conf.no_sync_fortran:
        _sync_file(metadata, conf.fortran_file, Fortran)

    if conf.pom_xml_file.is_file() and not conf.no_sync_pom_xml:
        _sync_file(metadata, conf.pom_xml_file, POM)

    if conf.mkdocs_file.is_file() and not conf.no_sync_mkdocs:
        _sync_file(metadata, conf.mkdocs_file, MkDocs)

    if conf.rust_file.is_file() and not conf.no_sync_rust:
        _sync_file(metadata, conf.rust_file, Rust)

    # create these by default if they are missing:
    if not conf.no_sync_cff:
        _sync_file(metadata, conf.cff_file, CFF)

    if not conf.no_sync_codemeta:
        _sync_file(metadata, conf.codemeta_file, CodeMeta)



"""Pyproject models."""
from typing import Optional, Set

from packaging.version import parse as parse_version
from pydantic import (
    BaseModel,
    Field,
    field_validator,
)
from typing_extensions import Annotated

from somesy.core.types import HttpUrlStr


class FortranConfig(BaseModel):
    """Fortran configuration model."""

    model_config = dict(use_enum_values=True)

    name: Annotated[
        str,
        Field(pattern=r"^[A-Za-z0-9]+([_-][A-Za-z0-9]+)*$", description="Package name"),
    ]
    version: Annotated[
        Optional[str],
        Field(
            pattern=r"^\d+(\.\d+)*((a|b|rc)\d+)?(post\d+)?(dev\d+)?$",
            description="Package version",
        ),
    ] = None
    description: Annotated[
        Optional[str], Field(description="Package description")
    ] = None
    license: Annotated[
        Optional[str],
        Field(description="SPDX license identifier(s)."),
    ] = None
    author: Annotated[
        Optional[str], Field(description="Package author information")
    ] = None
    maintainer: Annotated[
        Optional[str], Field(description="Package maintainer information")
    ] = None
    copyright: Annotated[
        Optional[str], Field(description="Package copyright text")
    ] = None
    homepage: Annotated[
        Optional[HttpUrlStr], Field(description="Package homepage")
    ] = None
    keywords: Annotated[
        Optional[Set[str]], Field(description="Keywords that describe the package")
    ] = None
    categories: Annotated[
        Optional[Set[str]], Field(description="Categories that package falls into")
    ] = None

    @field_validator("version")
    @classmethod
    def validate_version(cls, v):
        """Validate version using PEP 440."""
        try:
            _ = parse_version(v)
        except ValueError as err:
            raise ValueError("Invalid version") from err
        return v



"""Pyproject writers for setuptools and rust."""
import logging
from pathlib import Path
from typing import Any, List, Optional, Union

from rich.pretty import pretty_repr
from tomlkit import dump, load, table

from somesy.core.models import Person, ProjectMetadata
from somesy.core.writer import FieldKeyMapping, IgnoreKey, ProjectMetadataWriter

from .models import RustConfig, check_keyword

logger = logging.getLogger("somesy")


class Rust(ProjectMetadataWriter):
    """Rust config file handler parsed from Cargo.toml."""

    def __init__(self, path: Path):
        """Rust config file handler parsed from Cargo.toml.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        self._section = ["package"]
        mappings: FieldKeyMapping = {
            "maintainers": IgnoreKey(),
        }
        super().__init__(path, create_if_not_exists=False, direct_mappings=mappings)

    def _load(self) -> None:
        """Load Cargo.toml file."""
        with open(self.path) as f:
            self._data = load(f)

    def _validate(self) -> None:
        """Validate rust config using pydantic class.

        In order to preserve toml comments and structure, tomlkit library is used.
        Pydantic class only used for validation.
        """
        config = dict(self._get_property([]))
        logger.debug(
            f"Validating config using {RustConfig.__name__}: {pretty_repr(config)}"
        )
        RustConfig(**config)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the Cargo.toml file."""
        path = path or self.path
        with open(path, "w") as f:
            dump(self._data, f)

    def _get_property(
        self, key: Union[str, List[str], IgnoreKey], *, remove: bool = False, **kwargs
    ) -> Optional[Any]:
        """Get a property from the Cargo.toml file."""
        if isinstance(key, IgnoreKey):
            return None
        key_path = [key] if isinstance(key, str) else key
        full_path = self._section + key_path
        return super()._get_property(full_path, remove=remove, **kwargs)

    def _set_property(self, key: Union[str, List[str], IgnoreKey], value: Any) -> None:
        """Set a property in the Cargo.toml file."""
        if isinstance(key, IgnoreKey):
            return
        key_path = [key] if isinstance(key, str) else key

        if not value:  # remove value and clean up the sub-dict
            self._get_property(key_path, remove=True)
            return

        # get the tomlkit object of the section
        dat = self._get_property([])

        # dig down, create missing nested objects on the fly
        curr = dat
        for key in key_path[:-1]:
            if key not in curr:
                curr.add(key, table())
            curr = curr[key]
        curr[key_path[-1]] = value

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to rust string for person format "full name <email>."""
        return person.to_name_email_string()

    @staticmethod
    def _to_person(person_obj: str) -> Person:
        """Parse rust person string to a Person. It has format "full name <email>." but email is optional."""
        try:
            return Person.from_name_email_string(person_obj)
        except (ValueError, AttributeError):
            return None

    @classmethod
    def _parse_people(cls, people: Optional[List[Any]]) -> List[Person]:
        """Return a list of Persons parsed from list of format-specific people representations. to_person can return None, so filter out None values."""
        return list(filter(None, map(cls._to_person, people or [])))

    @property
    def keywords(self) -> Optional[List[str]]:
        """Return the keywords of the project."""
        return self._get_property(self._get_key("keywords"))

    @keywords.setter
    def keywords(self, keywords: List[str]) -> None:
        """Set the keywords of the project."""
        validated_keywords = []
        for keyword in keywords:
            try:
                check_keyword(keyword)
                validated_keywords.append(keyword)
            except ValueError as e:
                logger.debug(f"Invalid keyword {keyword}: {e}")

        # keyword count should max 5, so delete the rest
        if len(validated_keywords) > 5:
            validated_keywords = validated_keywords[:5]
        self._set_property(self._get_key("keywords"), validated_keywords)

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync the rust config with the project metadata."""
        super().sync(metadata)

        # if there is a license file, remove the license field
        if self._get_key("license_file"):
            self.license = None



"""Wrapper to provide dict-like access to XML via ElementTree."""
from __future__ import annotations

import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, List, Optional, Union, cast

import defusedxml.ElementTree as DET

# shallow type hint mostly for documentation purpose
JSONLike = Any


def load_xml(path: Path) -> ET.ElementTree:
    """Parse an XML file into an ElementTree, preserving comments."""
    path = path if isinstance(path, Path) else Path(path)
    parser = DET.XMLParser(target=ET.TreeBuilder(insert_comments=True))
    return DET.parse(path, parser=parser)


def indent(elem, level=0):
    """Indent the elements of this XML node (i.e. pretty print)."""
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for el in elem:
            indent(el, level + 1)
        if not el.tail or not el.tail.strip():
            el.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


class XMLProxy:
    """Class providing dict-like access to edit XML via ElementTree.

    Note that this wrapper facade is limited to a restricted (but useful) subset of XML:
    * XML attributes are not supported
    * DTDs are ignored (arbitrary keys can be queried and added)
    * each tag is assumed to EITHER contain text OR more nested tags
    * lists are treated atomically (no way to add/remove element from a collection)

    The semantics is implemented as follows:

    * If there are multiple tags with the same name, a list of XMLProxy nodes is returned
    * If a unique tag does have no nested tags, its `text` string value is returned
    * Otherwise, the node is returned
    """

    def _wrap(self, el: ET.Element) -> XMLProxy:
        """Wrap a different element, inheriting the same namespace."""
        return XMLProxy(el, default_namespace=self._def_ns)

    def _dump(self):
        """Dump XML to stdout (for debugging)."""
        ET.dump(self._node)

    def _qualified_key(self, key: str):
        """If passed key is not qualified, prepends the default namespace (if set)."""
        if key[0] == "{" or not self._def_ns:
            return key
        return "{" + self._def_ns + "}" + key

    def _shortened_key(self, key: str):
        """Inverse of `_qualified_key` (strips default namespace from element name)."""
        if key[0] != "{" or not self._def_ns or key.find(self._def_ns) < 0:
            return key
        return key[key.find("}") + 1 :]

    # ----

    def __init__(self, el: ET.Element, *, default_namespace: Optional[str] = None):
        """Wrap an existing XML ElementTree Element."""
        self._node: ET.Element = el
        self._def_ns = default_namespace

    @classmethod
    def parse(cls, path: Union[str, Path], **kwargs) -> XMLProxy:
        """Parse an XML file into a wrapped ElementTree, preserving comments."""
        path = path if isinstance(path, Path) else Path(path)
        return cls(load_xml(path).getroot(), **kwargs)

    def write(self, path: Union[str, Path], *, header: bool = True, **kwargs):
        """Write the XML DOM to an UTF-8 encoded file."""
        path = path if isinstance(path, Path) else Path(path)
        et = ET.ElementTree(self._node)
        if self._def_ns and "default_namespace" not in kwargs:
            kwargs["default_namespace"] = self._def_ns
        indent(et.getroot())
        et.write(path, encoding="UTF-8", xml_declaration=header, **kwargs)

    def __repr__(self):
        """See `object.__repr__`."""
        return str(self._node)

    def __len__(self):
        """Return number of inner tags inside current XML element.

        Note that bool(node) thus checks whether an XML node is a leaf in the element tree.
        """
        return len(self._node)

    def __iter__(self):
        """Iterate the nested elements in-order."""
        return map(self._wrap, iter(self._node))

    @property
    def namespace(self) -> Optional[str]:
        """Default namespace of this node."""
        return self._def_ns

    @property
    def is_comment(self):
        """Return whether the current element node is an XML comment."""
        return not isinstance(self._node.tag, str)

    @property
    def tag(self) -> Optional[str]:
        """Return tag name of this element (unless it is a comment)."""
        if self.is_comment:
            return None
        return self._shortened_key(self._node.tag)

    @tag.setter
    def tag(self, val: str):
        """Set the tag of this element."""
        if self.is_comment:
            raise ValueError("Cannot set tag name for comment element!")
        self._node.tag = self._qualified_key(val)

    # ---- helpers ----

    def to_jsonlike(
        self,
        *,
        strip_default_ns: bool = True,
        keep_root: bool = False,
    ) -> JSONLike:
        """Convert XML node to a JSON-like primitive, array or dict (ignoring attributes).

        Note that all leaf values are strings (i.e. not parsed to bool/int/float etc.).

        Args:
            strip_default_ns: Do not qualify keys from the default namespace
            keep_root: If true, the root tag name will be preserved (`{"root_tag": {...}}`)
        """
        if not len(self):  # leaf -> assume it's a primitive value
            return self._node.text or ""

        dct = {}
        ccnt = 0
        for elem in iter(self):
            raw = elem._node
            if not isinstance(raw.tag, str):
                ccnt += 1
                key = f"__comment_{ccnt}__"
            else:
                key = raw.tag if not strip_default_ns else self._shortened_key(raw.tag)

            curr_val = elem.to_jsonlike(strip_default_ns=strip_default_ns)
            if key not in dct:
                dct[key] = curr_val
                continue
            val = dct[key]
            if not isinstance(val, list):
                dct[key] = [dct[key]]
            dct[key].append(curr_val)

        return dct if not keep_root else {self._shortened_key(self._node.tag): dct}

    @classmethod
    def _from_jsonlike_primitive(
        cls, val, *, elem_name: Optional[str] = None, **kwargs
    ) -> Union[str, XMLProxy]:
        """Convert a leaf node into a string value (i.e. return inner text).

        Returns a string (or an XML element, if elem_name is passed).
        """
        if val is None:
            ret = ""  # turn None into empty string
        elif isinstance(val, str):
            ret = val
        elif isinstance(val, bool):
            ret = str(val).lower()  # True -> true / False -> false
        elif isinstance(val, (int, float)):
            ret = str(val)
        else:
            raise TypeError(
                f"Value of type {type(val)} is not JSON-like primitive: {val}"
            )

        if not elem_name:
            return ret
        else:  # return the value wrapped as an element (needed in from_jsonlike)
            elem = ET.Element(elem_name)
            elem.text = ret
            return cls(elem, **kwargs)

    @classmethod
    def from_jsonlike(
        cls, val: JSONLike, *, root_name: Optional[str] = None, **kwargs: Any
    ):
        """Convert a JSON-like primitive, array or dict into an XML element.

        Note that booleans are serialized as `true`/`false` and None as `null`.

        Args:
            val: Value to convert into an XML element.
            root_name: If `val` is a dict, defines the tag name for the root element.
            kwargs: Additional arguments for XML element instantiation.
        """
        if isinstance(val, list):
            return list(
                map(lambda x: cls.from_jsonlike(x, root_name=root_name, **kwargs), val)
            )
        if not isinstance(val, dict):  # primitive val
            return cls._from_jsonlike_primitive(val, elem_name=root_name, **kwargs)

        # now the dict case remains
        elem = ET.Element(root_name or "root")
        for k, v in val.items():
            if k.startswith(
                "__comment_"
            ):  # special key names are mapped to XML comments
                elem.append(ET.Comment(v if isinstance(v, str) else str(v)))

            elif isinstance(v, list):
                for vv in XMLProxy.from_jsonlike(v, root_name=k, **kwargs):
                    elem.append(vv._node)
            elif not isinstance(v, dict):  # primitive val
                # FIXME: use better case-splitting for type of function to avoid cast
                tmp = cast(
                    XMLProxy,
                    XMLProxy._from_jsonlike_primitive(v, elem_name=k, **kwargs),
                )
                elem.append(tmp._node)
            else:  # dict
                elem.append(XMLProxy.from_jsonlike(v, root_name=k)._node)

        return cls(elem, **kwargs)

    # ---- dict-like access ----

    def get(self, key: str, *, as_nodes: bool = False, deep: bool = False):
        """Get sub-structure(s) of value(s) matching desired XML tag name.

        * If there are multiple matching elements, will return them all as a list.
        * If there is a single matching element, will return that element without a list.

        Args:
            key: tag name to retrieve
            as_nodes: If true, will *always* return a list of (zero or more) XML nodes
            deep: Expand nested XML elements instead of returning them as XML nodes
        """
        # NOTE: could allow to retrieve comments when using empty string/none as key?

        if as_nodes and deep:
            raise ValueError("as_nodes and deep are mutually exclusive!")
        if not key:
            raise ValueError("Key must not be an empty string!")
        key = self._qualified_key(key)

        # if not fully qualified + default NS is given, use it for query
        lst = self._node.findall(key)
        ns: List[XMLProxy] = list(map(self._wrap, lst))
        if as_nodes:  # return it as a list of xml nodes
            return ns
        if not ns:  # no element
            return None

        ret = ns if not deep else [x.to_jsonlike() for x in ns]
        if len(ret) == 1:
            return ret[0]  # single element
        else:
            return ret

    def __getitem__(self, key: str):
        """Acts like `dict.__getitem__`, implemented with `get`."""
        val = self.get(key)
        if val is not None:
            return val
        else:
            raise KeyError(key)

    def __contains__(self, key: str) -> bool:
        """Acts like `dict.__contains__`, implemented with `get`."""
        return self.get(key) is not None

    def __delitem__(self, key: Union[str, XMLProxy]):
        """Delete a nested XML element with matching key name.

        Note that **all** XML elements with the given tag name are removed!

        To prevent this behavior, instead of a string tag name you can provide the
        exact element to be removed, i.e. if a node `node_a` represents the following XML:

        ```
        <a>
          <b>1</b>
          <c>2</c>
          <b>3</b>
        </a>
        ```

        Then we have that:

        * `del node_a["b"]` removes **both** tags, leaving just the `c` tag.
        * `del node_a[node_a["a"][1]]` removes just the second tag with the `3`.
        """
        if isinstance(key, str):
            nodes = self.get(key, as_nodes=True)
        else:
            nodes = [key] if key._node in self._node else []

        if not nodes:
            raise KeyError(key)

        if self._node.text is not None:
            self._node.text = ""
        for child in nodes:
            self._node.remove(child._node)

    def _clear(self):
        """Remove contents of this XML element (e.g. for overwriting in-place)."""
        self._node.text = ""
        children = list(iter(self._node))  # need to store, removal invalidates iterator
        for child in children:
            self._node.remove(child)

    def __setitem__(self, key: Union[str, XMLProxy], val: Union[JSONLike, XMLProxy]):
        """Add or overwrite an inner XML tag.

        If there is exactly one matching tag, the value is substituted in-place.
        If the passed value is a list, all list entries are added in their own element.

        If there are multiple existing matches or target values, then
        **all** existing elements are removed and the new value(s) are added in
        new element(s) (i.e. coming after other unrelated existing elements)!

        To prevent this behavior, instead of a string tag name you can provide the
        exact element to be overwritten, i.e. if a node `node_a` represents the following XML:

        ```
        <a>
          <b>1</b>
          <c>2</c>
          <b>3</b>
        </a>
        ```

        Then we have that:

        * `node_a["b"] = 5` removes both existing tags and creates a new tag with the passed value(s).
        * `node_a[node_a["b"][1]] = 5` replaces the `3` in the second tag with the `5`.

        Note that the passed value must be either an XML element already, or be a pure JSON-like object.
        """
        if isinstance(key, str):
            nodes = self.get(key, as_nodes=True)
            # delete all existing elements if multiple exist or are passed
            if len(nodes) > 1 or (len(nodes) and isinstance(val, list)):
                del self[key]
                nodes = []
            # now we can assume there's zero or one suitable target elements
            if nodes:  # if it is one, clear it out
                nodes[0]._clear()
        else:  # an XMLProxy object was passed as key -> try to use that
            if isinstance(val, list):
                raise ValueError(
                    "Cannot overwrite a single element with a list of values!"
                )
            # ensure the target node is cleared out and use it as target
            key._clear()
            nodes = [key]
            key = key.tag

        # ensure key string is qualified with a namespace
        key_name: str = self._qualified_key(key)

        # normalize passed value(s) to be list (general case)
        vals = val if isinstance(val, list) else [val]

        # ensure there is the required number of target element nodes
        for _ in range(len(vals) - len(nodes)):
            nodes.append(self._wrap(ET.SubElement(self._node, key_name)))

        # normalize values no XML element nodes
        nvals = []
        for val in vals:
            # ensure value is represented as an XML node
            if isinstance(val, XMLProxy):
                obj = self._wrap(ET.Element("dummy"))
                obj._node.append(val._node)
            else:
                obj = self.from_jsonlike(val, root_name=key_name)

            nvals.append(obj)

        for node, val in zip(nodes, nvals):
            # transplant node contents into existing element (so it is inserted in-place)
            node._node.text = val._node.text
            for child in iter(val):
                node._node.append(child._node)



import pytest
from tomlkit import dump

from somesy.fortran import Fortran


def test_fortran_validate_accept(load_files, file_types):
    """Validate by loading the data fpm.toml file using the fixture."""
    load_files([file_types.FORTRAN])


def test_fortran_validate(tmp_path):
    """Test validating a fpm.toml file."""

    # create a fpm.toml file but with a invalid values
    reject_fortran_object = {
        "name": "12somesy",
        "version": "abc",
        "authors": ["John Doe <"],
    }

    invalid_fortran_path = tmp_path / "fpm.toml"
    with open(invalid_fortran_path, "w+") as f:
        dump(reject_fortran_object, f)

    with pytest.raises(ValueError):
        Fortran(invalid_fortran_path)



"""Pyproject models."""
from typing import Optional

from pydantic import (
    BaseModel,
    Field,
)
from typing_extensions import Annotated

from somesy.core.types import HttpUrlStr


class MkDocsConfig(BaseModel):
    """MkDocs configuration model."""

    model_config = dict(use_enum_values=True)

    site_name: Annotated[
        str,
        Field(pattern=r"^[A-Za-z0-9]+([_-][A-Za-z0-9]+)*$", description="Site name"),
    ]
    site_description: Annotated[str, Field(description="Site description")] = None
    site_author: Annotated[str, Field(description="Site authors")] = None
    site_url: Annotated[Optional[HttpUrlStr], Field(description="Site homepage")] = None
    repo_url: Annotated[
        Optional[HttpUrlStr], Field(description="Package repository")
    ] = None
    repo_name: Annotated[str, Field(description="Repository name")] = None



"""Utility functions for somesy."""
import json

import wrapt


@wrapt.decorator
def json_dump_wrapper(wrapped, instance, args, kwargs):
    """Wrap json.dump to write non-ascii characters with default indentation."""
    # Ensure ensure_ascii is set to False
    kwargs["ensure_ascii"] = False
    # set indent to 2 if not set
    kwargs["indent"] = kwargs.get("indent", 2)
    return wrapped(*args, **kwargs)


# Apply the wrapper
json.dump = json_dump_wrapper(json.dump)



"""Julia writer."""
import logging
from pathlib import Path
from typing import Optional

import tomlkit
from rich.pretty import pretty_repr

from somesy.core.models import Person, ProjectMetadata
from somesy.core.writer import ProjectMetadataWriter

from .models import JuliaConfig

logger = logging.getLogger("somesy")


class Julia(ProjectMetadataWriter):
    """Julia config file handler parsed from Project.toml."""

    def __init__(self, path: Path):
        """Julia config file handler parsed from Project.toml.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        super().__init__(path, create_if_not_exists=False)

    def _load(self) -> None:
        """Load Project.toml file."""
        with open(self.path) as f:
            self._data = tomlkit.load(f)

    def _validate(self) -> None:
        """Validate poetry config using pydantic class.

        In order to preserve toml comments and structure, tomlkit library is used.
        Pydantic class only used for validation.
        """
        config = dict(self._get_property([]))
        logger.debug(
            f"Validating config using {JuliaConfig.__name__}: {pretty_repr(config)}"
        )
        JuliaConfig(**config)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the julia file."""
        path = path or self.path
        with open(path, "w") as f:
            tomlkit.dump(self._data, f)

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to a name+email string."""
        return person.to_name_email_string()

    @staticmethod
    def _to_person(person_obj: str) -> Person:
        """Parse name+email string to a Person."""
        return Person.from_name_email_string(person_obj)

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync output file with other metadata files."""
        # overridden to not sync fields that are not present in the Project.toml file
        self.name = metadata.name
        self.version = metadata.version

        self._sync_authors(metadata)



"""Pyproject models."""
import re
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from packaging.version import parse as parse_version
from pydantic import BaseModel, Field, field_validator, model_validator
from typing_extensions import Annotated

from somesy.core.types import HttpUrlStr


class RustConfig(BaseModel):
    """Rust configuration model."""

    model_config = dict(use_enum_values=True)

    name: Annotated[
        str,
        Field(
            pattern=r"^[A-Za-z0-9]+([_-][A-Za-z0-9]+)*$",
            max_length=64,
            description="Package name",
        ),
    ]
    version: Annotated[
        str,
        Field(
            pattern=r"^\d+(\.\d+)*((a|b|rc)\d+)?(post\d+)?(dev\d+)?$",
            description="Package version",
        ),
    ]
    description: Annotated[
        Optional[str], Field(description="Package description")
    ] = None
    license: Annotated[
        Optional[str],
        Field(
            description="A combination SPDX license identifiers with AND, OR and so on."
        ),
    ] = None
    authors: Annotated[Set[str], Field(description="Package authors")]
    maintainers: Annotated[
        Optional[Set[str]], Field(description="Package maintainers")
    ] = None
    readme: Annotated[
        Optional[Union[Path, List[Path]]], Field(description="Package readme file(s)")
    ] = None
    license_file: Annotated[
        Optional[Path], Field(description="Package license file")
    ] = None
    homepage: Annotated[
        Optional[HttpUrlStr], Field(description="Package homepage")
    ] = None
    repository: Annotated[
        Optional[HttpUrlStr], Field(description="Package repository")
    ] = None
    documentation: Annotated[
        Optional[HttpUrlStr], Field(description="Package documentation page")
    ] = None
    keywords: Annotated[
        Optional[Set[str]], Field(description="Keywords that describe the package")
    ] = None
    classifiers: Annotated[
        Optional[List[str]], Field(description="pypi classifiers")
    ] = None
    urls: Annotated[
        Optional[Dict[str, HttpUrlStr]], Field(description="Package URLs")
    ] = None

    @model_validator(mode="before")
    @classmethod
    def license_or_file(cls, values):
        """License and license file are mutually exclusive."""
        if "license" in values and "license_file" in values:
            raise ValueError("license and license_file are mutually exclusive")
        return values

    @field_validator("version")
    @classmethod
    def validate_version(cls, v):
        """Validate version using PEP 440."""
        try:
            _ = parse_version(v)
        except ValueError as err:
            raise ValueError("Invalid version") from err
        return v

    @field_validator("readme", "license_file")
    @classmethod
    def validate_readme(cls, v):
        """Validate readme file(s) by checking whether files exist."""
        if isinstance(v, list):
            if any(not e.is_file() for e in v):
                raise ValueError("Some file(s) do not exist")
        else:
            if not v.is_file():
                raise ValueError("File does not exist")

    @field_validator("keywords")
    @classmethod
    def check_keywords_field(cls, v):
        """Check the keywords field."""
        if v is None:
            return v

        # Check if number of keywords is at most 5
        if v is not None and len(v) > 5:
            raise ValueError("A maximum of 5 keywords is allowed")

        for keyword in v:
            check_keyword(keyword)

        return v


def check_keyword(keyword: str):
    """Check if keyword is valid."""
    # Check if keyword is ASCII and has at most 20 characters
    if not keyword.isascii() or len(keyword) > 20:
        raise ValueError(
            "Each keyword must be ASCII text and have at most 20 characters"
        )

    # Check if keyword starts with an alphanumeric character
    if not re.match(r"^[a-zA-Z0-9]", keyword):
        raise ValueError("Each keyword must start with an alphanumeric character")

    # Check if keyword contains only allowed characters
    if not re.match(r"^[a-zA-Z0-9_\-+]+$", keyword):
        raise ValueError("Keywords can only contain letters, numbers, _, -, or +")



"""Fortran writer."""
import logging
from pathlib import Path
from typing import Any, List, Optional

import tomlkit
from rich.pretty import pretty_repr

from somesy.core.models import Person, ProjectMetadata
from somesy.core.writer import FieldKeyMapping, IgnoreKey, ProjectMetadataWriter

from .models import FortranConfig

logger = logging.getLogger("somesy")


class Fortran(ProjectMetadataWriter):
    """Fortran config file handler parsed from fpm.toml."""

    def __init__(self, path: Path):
        """Fortran config file handler parsed from fpm.toml.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        mappings: FieldKeyMapping = {
            "authors": ["author"],
            "maintainers": ["maintainer"],
            "documentation": IgnoreKey(),
        }
        super().__init__(path, create_if_not_exists=False, direct_mappings=mappings)

    @property
    def authors(self):
        """Return the only author of the fpm.toml file as list."""
        authors = []
        try:
            self._to_person(self._get_property(self._get_key("authors")))
            authors = [self._get_property(self._get_key("authors"))]
        except ValueError:
            logger.warning("Cannot convert authors to Person object.")
        return authors

    @authors.setter
    def authors(self, authors: List[Person]) -> None:
        """Set the authors of the project."""
        authors = self._from_person(authors[0])
        self._set_property(self._get_key("authors"), authors)

    @property
    def maintainers(self):
        """Return the only author of the fpm.toml file as list."""
        maintainers = self._get_property(self._get_key("maintainers"))
        if maintainers:
            return [self._get_property(self._get_key("maintainers"))]
        return []

    @maintainers.setter
    def maintainers(self, maintainers: List[Person]) -> None:
        """Set the maintainers of the project."""
        maintainers = self._from_person(maintainers[0])
        self._set_property(self._get_key("maintainers"), maintainers)

    def _load(self) -> None:
        """Load fpm.toml file."""
        with open(self.path) as f:
            self._data = tomlkit.load(f)

    def _validate(self) -> None:
        """Validate poetry config using pydantic class.

        In order to preserve toml comments and structure, tomlkit library is used.
        Pydantic class only used for validation.
        """
        config = dict(self._get_property([]))
        logger.debug(
            f"Validating config using {FortranConfig.__name__}: {pretty_repr(config)}"
        )
        FortranConfig(**config)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the fpm file."""
        path = path or self.path
        with open(path, "w") as f:
            tomlkit.dump(self._data, f)

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to poetry string for person format "full name <email>."""
        return person.to_name_email_string()

    @staticmethod
    def _to_person(person_obj: Any) -> Person:
        """Cannot convert from free string to person object."""
        try:
            return Person.from_name_email_string(person_obj)
        except ValueError:
            logger.warning(f"Cannot convert {person_obj} to Person object.")
            return None

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync output file with other metadata files."""
        self.name = metadata.name
        self.description = metadata.description

        if metadata.version:
            self.version = metadata.version

        if metadata.keywords:
            self.keywords = metadata.keywords

        self.authors = metadata.authors()
        maintainers = metadata.maintainers()

        # set if not empty
        if maintainers:
            # only one maintainer is allowed
            self.maintainers = maintainers

        self.license = metadata.license.value

        self.homepage = str(metadata.homepage) if metadata.homepage else None



"""codemeta.json creation module."""
import logging
from collections import OrderedDict
from pathlib import Path
from typing import Any, List, Optional

from rich.pretty import pretty_repr

from somesy.core.models import Person, ProjectMetadata
from somesy.core.writer import FieldKeyMapping, ProjectMetadataWriter
from somesy.json_wrapper import json

logger = logging.getLogger("somesy")


class CodeMeta(ProjectMetadataWriter):
    """Codemeta.json parser and saver."""

    def __init__(
        self,
        path: Path,
    ):
        """Codemeta.json parser.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        mappings: FieldKeyMapping = {
            "repository": ["codeRepository"],
            "homepage": ["softwareHelp"],
            "documentation": ["buildInstructions"],
            "keywords": ["keywords"],
            "authors": ["author"],
            "maintainers": ["maintainer"],
            "contributors": ["contributor"],
        }
        # delete the file if it exists
        if path.is_file():
            logger.verbose("Deleting existing codemeta.json file.")
            path.unlink()
        super().__init__(path, create_if_not_exists=True, direct_mappings=mappings)

    @property
    def authors(self):
        """Return the only author of the codemeta.json file as list."""
        return self._get_property(self._get_key("publication_authors")) or []

    @authors.setter
    def authors(self, authors: List[Person]) -> None:
        """Set the authors of the project."""
        authors = [self._from_person(a) for a in authors]
        self._set_property(self._get_key("authors"), authors)

    @property
    def contributors(self):
        """Return the contributors of the codemeta.json file."""
        return self._get_property(self._get_key("contributors"))

    @contributors.setter
    def contributors(self, contributors: List[Person]) -> None:
        """Set the contributors of the project."""
        contributors = [self._from_person(c) for c in contributors]
        self._set_property(self._get_key("contributors"), contributors)

    def _load(self) -> None:
        """Load codemeta.json file."""
        with self.path.open() as f:
            self._data = json.load(f, object_pairs_hook=OrderedDict)

    def _validate(self) -> None:
        """Validate codemeta.json content using pydantic class."""
        config = dict(self._get_property([]))

        logger.debug(
            f"No validation for codemeta.json files {CodeMeta.__name__}: {pretty_repr(config)}"
        )

    def _init_new_file(self) -> None:
        data = {
            "@context": [
                "https://doi.org/10.5063/schema/codemeta-2.0",
                "https://w3id.org/software-iodata",
                "https://raw.githubusercontent.com/jantman/repostatus.org/master/badges/latest/ontology.jsonld",
                "https://schema.org",
                "https://w3id.org/software-types",
            ],
            "@type": "SoftwareSourceCode",
            "author": [],
        }
        # dump to file
        with self.path.open("w+") as f:
            json.dump(data, f)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the codemeta.json file."""
        path = path or self.path
        logger.debug(f"Saving codemeta.json to {path}")

        # copy the _data
        data = self._data.copy()

        # set license
        if "license" in data:
            data["license"] = (f"https://spdx.org/licenses/{data['license']}",)

        # if softwareHelp is set, set url to softwareHelp
        if "softwareHelp" in data:
            data["url"] = data["softwareHelp"]

        with path.open("w") as f:
            # codemeta.json indentation is 2 spaces
            json.dump(data, f)

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to codemeta.json dict for person format."""
        person_dict = {
            "@type": "Person",
        }
        if person.given_names:
            person_dict["givenName"] = person.given_names
        if person.family_names:
            person_dict["familyName"] = person.family_names
        if person.email:
            person_dict["email"] = person.email
        if person.orcid:
            person_dict["@id"] = str(person.orcid)
        if person.address:
            person_dict["address"] = person.address
        if person.affiliation:
            person_dict["affiliation"] = person.affiliation
        return person_dict

    @staticmethod
    def _to_person(person) -> Person:
        """Convert codemeta.json dict or str for person format to project metadata person object."""
        person_obj = {}
        if "givenName" in person:
            person_obj["given_names"] = person["givenName"].strip()
        if "familyName" in person:
            person_obj["family_names"] = person["familyName"].strip()
        if "email" in person:
            person_obj["email"] = person["email"].strip()
        if "@id" in person:
            person_obj["orcid"] = person["@id"].strip()
        if "address" in person:
            person_obj["address"] = person["address"].strip()

        return Person(**person_obj)

    def _sync_person_list(self, old: List[Any], new: List[Person]) -> List[Any]:
        """Override the _sync_person_list function from ProjectMetadataWriter.

        This method wont care about existing persons in codemeta.json file.

        Args:
            old (List[Any]): _description_
            new (List[Person]): _description_

        Returns:
            List[Any]: _description_
        """
        return new

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync codemeta.json with project metadata.

        Use existing sync function from ProjectMetadataWriter but update repository and contributors.
        """
        super().sync(metadata)
        self.contributors = self._sync_person_list(
            self.contributors, metadata.contributors()
        )



import pytest
from tomlkit import dump

from somesy.julia import Julia


def test_julia_validate_accept(load_files, file_types):
    """Validate by loading the data Project.toml file using the fixture."""
    load_files([file_types.JULIA])


def test_julia_validate(tmp_path):
    """Test validating a Project.toml file."""

    # create a Project.toml file but with a invalid values
    reject_julia_object = {
        "name": "somesy",
        "version": "abc",
        "authors": ["John Doe <"],
    }

    invalid_julia_path = tmp_path / "Project.toml"
    with open(invalid_julia_path, "w+") as f:
        dump(reject_julia_object, f)

    with pytest.raises(ValueError):
        Julia(invalid_julia_path)



"""Project documentation with Markdown (MkDocs) parser and saver."""
import logging
from pathlib import Path
from typing import List, Optional

from rich.pretty import pretty_repr
from ruamel.yaml import YAML

from somesy.core.models import Person, ProjectMetadata
from somesy.core.writer import FieldKeyMapping, IgnoreKey, ProjectMetadataWriter
from somesy.mkdocs.models import MkDocsConfig

logger = logging.getLogger("somesy")


class MkDocs(ProjectMetadataWriter):
    """Project documentation with Markdown (MkDocs) parser and saver."""

    def __init__(self, path: Path, create_if_not_exists: bool = False):
        """Project documentation with Markdown (MkDocs) parser.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        self._yaml = YAML()
        self._yaml.preserve_quotes = True

        mappings: FieldKeyMapping = {
            "name": ["site_name"],
            "description": ["site_description"],
            "homepage": ["site_url"],
            "repository": ["repo_url"],
            "authors": ["site_author"],
            "documentation": IgnoreKey(),
            "version": IgnoreKey(),
            "maintainers": IgnoreKey(),
            "license": IgnoreKey(),
            "keywords": IgnoreKey(),
        }
        super().__init__(
            path, create_if_not_exists=create_if_not_exists, direct_mappings=mappings
        )

    def _load(self):
        """Load the MkDocs file."""
        with open(self.path) as f:
            self._data = self._yaml.load(f)

    def _validate(self):
        """Validate the MkDocs file."""
        config = dict(self._get_property([]))
        logger.debug(
            f"Validating config using {MkDocsConfig.__name__}: {pretty_repr(config)}"
        )
        MkDocsConfig(**config)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the MkDocs object to a file."""
        path = path or self.path
        self._yaml.dump(self._data, path)

    @property
    def authors(self):
        """Return the only author from the source file as list."""
        authors = []
        try:
            self._to_person(self._get_property(self._get_key("authors")))
            authors = [self._get_property(self._get_key("authors"))]
        except AttributeError:
            logger.warning("Cannot convert authors to Person object.")
        return authors

    @authors.setter
    def authors(self, authors: List[Person]) -> None:
        """Set the authors of the project."""
        authors = self._from_person(authors[0])
        self._set_property(self._get_key("authors"), authors)

    @staticmethod
    def _from_person(person: Person):
        """MkDocs Person is a string with full name."""
        return person.to_name_email_string()

    @staticmethod
    def _to_person(person: str):
        """MkDocs Person is a string with full name."""
        return Person.from_name_email_string(person)

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync the MkDocs object with the ProjectMetadata object."""
        self.name = metadata.name
        self.description = metadata.description
        # no author merge since it is a free text field
        self.authors = metadata.authors()
        if metadata.homepage:
            self.homepage = str(metadata.homepage)
        if metadata.repository:
            self.repository = str(metadata.repository)
            self.repo_name = metadata.repository.path



"""Somesy implementation for Java Maven pom.xml."""

# some POM-related constants and reusable objects
POM_URL = "http://maven.apache.org/POM/4.0.0"
POM_PREF = "{" + POM_URL + "}"
POM_ROOT_ATRS = {
    "xmlns": "http://maven.apache.org/POM/4.0.0",
    "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "xsi:schemaLocation": "http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd",
}



from pathlib import Path

import pytest

from somesy.core.models import LicenseEnum, Person, ProjectMetadata
from somesy.fortran.writer import Fortran


@pytest.fixture
def fortran(load_files, file_types):
    files = load_files([file_types.FORTRAN])
    return files[file_types.FORTRAN]


@pytest.fixture
def fortran_file(create_files, file_types):
    folder = create_files([(file_types.FORTRAN, "fpm.toml")])
    return folder / Path("fpm.toml")


def test_content_match(fortran):
    assert fortran.name == "test-package"
    assert len(fortran.authors) == 1


def test_sync(fortran, somesy_input):
    # with 'full name <email>' format
    fortran.sync(somesy_input.project)
    assert fortran.name == "testproject"
    assert fortran.version == "1.0.0"
    assert fortran.authors[0] == "John Doe <john.doe@example.com>"


def test_sync_free_text(fortran_file, somesy_input):
    # update author to have a free text format
    content = fortran_file.read_text()
    content = content.replace(
        'author = "John Doe <john.doe@example.com>"', 'author = "John Doe"'
    )

    # write the new content
    fortran_file.write_text(content)

    # read the file again
    fortran = Fortran(fortran_file)
    fortran.sync(somesy_input.project)
    assert fortran.name == "testproject"
    assert fortran.version == "1.0.0"
    assert fortran.authors[0] == "John Doe <john.doe@example.com>"


def test_save(tmp_path, fortran):
    custom_path = tmp_path / Path("fpm.toml")
    fortran.save(custom_path)
    assert custom_path.is_file()
    custom_path.unlink()


def test_from_person(person):
    assert Fortran._from_person(person) == f"{person.full_name} <{person.email}>"



from pydantic import ValidationError
import pytest
from tomlkit import dump

from somesy.rust import Rust


def test_rust_validate_accept(load_files, file_types):
    """Validate by loading the data Cargo.toml file using the fixture."""
    load_files([file_types.RUST])


def test_rust_validate(tmp_path):
    """Test validating a Cargo.toml file."""

    # create a Cargo.toml file but with a invalid version value
    reject_rust_object = {
        "package": {
            "name": "somesy",
            "version": "abc",
            "authors": ["John Doe <"],
        }
    }

    invalid_rust_path = tmp_path / "Cargo.toml"
    with open(invalid_rust_path, "w+") as f:
        dump(reject_rust_object, f)

    with pytest.raises(ValidationError):
        Rust(invalid_rust_path)

    # reject with invalid values
    _reject_with(tmp_path, "name", "1test-")
    _reject_with(tmp_path, "keywords", ["testtesttesttesttesttesttest"])
    _reject_with(tmp_path, "keywords", ["test test"])
    _reject_with(tmp_path, "keywords", ["test*test"])
    _reject_with(
        tmp_path, "keywords", ["test", "test2", "test3", "test4", "test5", "test6"]
    )
    _reject_with(tmp_path, "license_file", "LICENSE")


def _reject_with(tmp_path, key, value):
    """Helper function to reject invalid values."""
    reject_rust_object = {
        "package": {
            "name": "somesy",
            "version": "1.0.0",
            "license": "MIT",
        }
    }
    reject_rust_object["package"][key] = value
    invalid_rust_path = tmp_path / "Cargo.toml"
    with open(invalid_rust_path, "w+") as f:
        dump(reject_rust_object, f)

    with pytest.raises(ValueError):
        Rust(invalid_rust_path)



from somesy.codemeta import CodeMeta
from somesy.json_wrapper import json


def test_update_codemeta(somesy_input, tmp_path):
    codemeta_file = tmp_path / "codemeta.json"

    cm = CodeMeta(codemeta_file)

    # firstly, create a codemeta file
    cm.sync(somesy_input.project)
    cm.save()

    assert codemeta_file.is_file()
    dat = open(codemeta_file, "rb").read()

    # second time, no changes but codemeta.json exists -> codemeta.json is the same
    cm.sync(somesy_input.project)
    cm.save()
    assert codemeta_file.is_file()
    dat2 = open(codemeta_file, "rb").read()
    assert dat == dat2

    # third time, change the project name -> codemeta.json is different
    somesy_input.project.name = "new_name"
    cm.sync(somesy_input.project)
    cm.save()
    assert codemeta_file.is_file()
    dat3 = open(codemeta_file, "rb").read()
    assert dat != dat3

    # fourth time, change the project name back and change version in codemeta.json
    # -> codemeta.json is the same
    somesy_input.project.name = "testproject"
    with open(codemeta_file, "w") as f:
        codemeta = json.loads(dat3)
        codemeta["version"] = "0.0.2"
        json.dump(codemeta, f)
    cm.sync(somesy_input.project)
    cm.save()
    assert codemeta_file.is_file()
    dat4 = open(codemeta_file, "rb").read()
    assert dat == dat4
    assert dat3 != dat4
